/*
 * wxcmisc.cc
 * wxWindows miscellaneous command processing
 *
 */

// For compilers that support precompilation, includes "wx.h".
#include "wx_prec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx.h"
#endif

#include "wx_hash.h"
#include "wx_types.h"
#include "wx_ipc.h"
#include "wx_timer.h"

#include <iostream.h>
#include <ctype.h>
#include <stdlib.h>

#include "wxextend.h"

/************************************************************************
 * DDE Classes
 *
 ************************************************************************/

// wxConnection
class wxExtensionConnection: public wxConnection
{
 public:
  wxExtensionConnection(void)
  {
    wxStoreThingInTable((long)this, this, wxTYPE_DDE_CONNECTION);
  }

 ~wxExtensionConnection(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_CONNECTION, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);

   wxDeleteThingInTable((long)this);
 }

 Bool OnAdvise(char *topic, char *item, char *data, int size, int format);

 Bool OnExecute(char *topic, char *data, int size, int format);

 Bool OnPoke(char *topic, char *item, char *data, int size, int format);

 char *OnRequest(char *topic, char *item, int *size, int format);

 Bool OnStartAdvise(char *topic, char *item);

 Bool OnStopAdvise(char *topic, char *item);
};

Bool wxExtensionConnection::OnAdvise(char *topic, char *item, char *data, int size, int format)
{
  wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_CONNECTION, "OnAdvise");
  if (func)
  {
    Bool handled = FALSE;
    Bool returnValue = (* (Bool (*)(Bool *, long, char *, char *, char *, int, int)) func)
     (&handled, (long)this, topic, item, data, size, format);

    if (!handled)
      return wxConnection::OnAdvise(topic, item, data, size, format);
    else
      return returnValue;
  }
  else
    return wxConnection::OnAdvise(topic, item, data, size, format);
}

Bool wxExtensionConnection::OnExecute(char *topic, char *data, int size, int format)
{
  wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_CONNECTION, "OnExecute");
  if (func)
  {
    Bool handled = FALSE;
    Bool returnValue = (* (Bool (*)(Bool *, long, char *, char *, int, int)) func)
     (&handled, (long)this, topic, data, size, format);

    if (!handled)
      return wxConnection::OnExecute(topic, data, size, format);
    else
      return returnValue;
  }
  else
    return wxConnection::OnExecute(topic, data, size, format);
}

Bool wxExtensionConnection::OnPoke(char *topic, char *item, char *data, int size, int format)
{
  wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_CONNECTION, "OnPoke");
  if (func)
  {
    Bool handled = FALSE;
    Bool returnValue = (* (Bool (*)(Bool *, long, char *, char *, char *, int, int)) func)
     (&handled, (long)this, topic, item, data, size, format);

    if (!handled)
      return wxConnection::OnPoke(topic, item, data, size, format);
    else
      return returnValue;
  }
  else
    return wxConnection::OnPoke(topic, item, data, size, format);
}

char *wxExtensionConnection::OnRequest(char *topic, char *item, int *size, int format)
{
  wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_CONNECTION, "OnRequest");
  if (func)
  {
    Bool handled = FALSE;
    char *returnValue = (* (char * (*)(Bool *, long, char *, char *, int *, int)) func)
     (&handled, (long)this, topic, item, size, format);

    if (!handled)
      return wxConnection::OnRequest(topic, item, size, format);
    else
      return returnValue;
  }
  else
    return wxConnection::OnRequest(topic, item, size, format);
}

Bool wxExtensionConnection::OnStartAdvise(char *topic, char *item)
{
  wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_CONNECTION, "OnStartAdvise");
  if (func)
  {
    Bool handled = FALSE;
    Bool returnValue = (* (Bool (*)(Bool *, long, char *, char *)) func)
     (&handled, (long)this, topic, item);

    if (!handled)
      return wxConnection::OnStartAdvise(topic, item);
    else
      return returnValue;
  }
  else
    return wxConnection::OnStartAdvise(topic, item);
}

Bool wxExtensionConnection::OnStopAdvise(char *topic, char *item)
{
  wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_CONNECTION, "OnStopAdvise");
  if (func)
  {
    Bool handled = FALSE;
    Bool returnValue = (* (Bool (*)(Bool *, long, char *, char *)) func)
     (&handled, (long)this, topic, item);

    if (!handled)
      return wxConnection::OnStopAdvise(topic, item);
    else
      return returnValue;
  }
  else
    return wxConnection::OnStopAdvise(topic, item);
}

// wxServer
class wxExtensionServer: public wxServer
{
 public:
  wxExtensionFunction clipsFunc;
  wxExtensionServer(char *service)
  {
    wxStoreThingInTable((long)this, this, wxTYPE_DDE_SERVER);
    Create(service);
  }

  ~wxExtensionServer(void)
  {
    wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_SERVER, "OnDelete");
    if (func)
      (* (void (*)(long)) func) ((long)this);

    wxDeleteThingInTable((long)this);
  }

  wxConnection *OnAcceptConnection(char *topic);
};

/*
 * Note that this changes how DDE servers work.
 * When the OnAcceptConnecton handler is called, the
 * connection ID is passed as well: if the application wishes to
 * accept the connection, it registers event handlers for this
 * connection and returns TRUE. Otherwise, it returns FALSE,
 * and the connection is deleted.
 */
 
wxConnection *wxExtensionServer::OnAcceptConnection(char *topic)
{
  wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_SERVER, "OnAcceptConnection");
  if (func)
  {
    Bool handled = FALSE;
    wxExtensionConnection *connection = new wxExtensionConnection;
    Bool returnValue = (* (Bool (*)(Bool *, long, char *, long)) func)
     (&handled, (long)this, topic, (long)connection);

    if (!handled || !returnValue)
    {
      delete connection;
      return NULL;
    }
    else
    {
      return connection;
    }
  }
  else
    return NULL;
}

// wxClient
class wxExtensionClient: public wxClient
{
 public:
  wxExtensionClient(void)
  {
    wxStoreThingInTable((long)this, this, wxTYPE_DDE_CLIENT);
  }

  ~wxExtensionClient(void)
  {
    wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_DDE_CLIENT, "OnDelete");
    if (func)
      (* (void (*)(long)) func) ((long)this);

    wxDeleteThingInTable((long)this);
  }

  wxConnection *OnMakeConnection(void)
  {
    wxExtensionConnection *connection = new wxExtensionConnection;
    return connection;
  }
};

/*
 * DDE functions
 *
 */

// DDE functions
WXID wxcClientCreate(void)
{
  wxExtensionClient *client = new wxExtensionClient;
  return (long)client;
}

WXID wxcClientMakeConnection(WXID id, char *host, char *service, char *topic)
{
  wxExtensionClient *client = (wxExtensionClient *)wxGetTypedObject(id, wxTYPE_DDE_CLIENT);
  if (!client)
  {
    ExtensionError << "Error in client-make-connection: could not find client " << id << ".\n";
    return 0;
  }

  wxConnection *connection = client->MakeConnection(host, service, topic);
  return (long)connection;
}

WXID wxcServerCreate(char *service)
{
  wxExtensionServer *server = new wxExtensionServer(service);
  return (long)server;
}

Bool wxcConnectionAdvise(WXID id, char *item, char *data)
{
  wxExtensionConnection *connection = (wxExtensionConnection *)wxGetTypedObject(id, wxTYPE_DDE_CONNECTION);
  if (!connection)
  {
    ExtensionError << "Error in connection-advise: could not find connection " << id << ".\n";
    return FALSE;
  }

  return connection->Advise(item, data);
}

Bool wxcConnectionExecute(WXID id, char *data)
{
  wxExtensionConnection *connection = (wxExtensionConnection *)wxGetTypedObject(id, wxTYPE_DDE_CONNECTION);
  if (!connection)
  {
    ExtensionError << "Error in connection-execute: could not find connection " << id << ".\n";
    return FALSE;
  }

  return connection->Execute(data);
}

Bool wxcConnectionDisconnect(WXID id)
{
  wxExtensionConnection *connection = (wxExtensionConnection *)wxGetTypedObject(id, wxTYPE_DDE_CONNECTION);
  if (!connection)
  {
    ExtensionError << "Error in connection-disconnect: could not find connection " << id << ".\n";
    return FALSE;
  }

  return connection->Disconnect();
}

Bool wxcConnectionPoke(WXID id, char *item, char *data)
{
  wxExtensionConnection *connection = (wxExtensionConnection *)wxGetTypedObject(id, wxTYPE_DDE_CONNECTION);
  if (!connection)
  {
    ExtensionError << "Error in connection-poke: could not find connection " << id << ".\n";
    return FALSE;
  }

  return connection->Poke(item, data);
}

char *wxcConnectionRequest(WXID id, char *item)
{
  wxExtensionConnection *connection = (wxExtensionConnection *)wxGetTypedObject(id, wxTYPE_DDE_CONNECTION);
  if (!connection)
  {
    ExtensionError << "Error in connection-request: could not find connection " << id << ".\n";
    return NULL;
  }

  char *data = connection->Request(item);
  return data;
}

Bool wxcConnectionStartAdvise(WXID id, char *item)
{
  wxExtensionConnection *connection = (wxExtensionConnection *)wxGetTypedObject(id, wxTYPE_DDE_CONNECTION);
  if (!connection)
  {
    ExtensionError << "Error in connection-start-advise: could not find connection " << id << ".\n";
    return FALSE;
  }

  return connection->StartAdvise(item);
}

Bool wxcConnectionStopAdvise(WXID id, char *item)
{
  wxExtensionConnection *connection = (wxExtensionConnection *)wxGetTypedObject(id, wxTYPE_DDE_CONNECTION);
  if (!connection)
  {
    ExtensionError << "Error in connection-stop-advise: could not find connection " << id << ".\n";
    return FALSE;
  }

  return connection->StopAdvise(item);
}
