/*
 * wxitemsp.h
 * wxCLIPS item classes
 *
 */

#ifndef exitemsph
#define exitemsph


// Button
class wxExtensionButton: public wxButton
{
 public:
  wxExtensionButton(wxPanel *panel, wxFunction func, char *label, int x = -1, int y = -1,
           int width = -1, int height = -1, long style = 0, char *name = "button"):
    wxButton(panel, func, label, x, y, width, height, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_BUTTON);
 }
  wxExtensionButton(wxPanel *panel, wxFunction func, wxBitmap *bitmap, int x = -1, int y = -1,
           int width = -1, int height = -1, long style = 0, char *name = "button"):
   wxButton(panel, func, bitmap, x, y, width, height, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_BUTTON);
 }
 ~wxExtensionButton(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_BUTTON, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};

// Group box
class wxExtensionGroupBox: public wxGroupBox
{
 public:
  wxExtensionGroupBox(wxPanel *panel, char *label, int x = -1, int y = -1,
           int width = -1, int height = -1, long style = 0, char *name = "groupBox"):
    wxGroupBox(panel, label, x, y, width, height, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_GROUP_BOX);
 }
 ~wxExtensionGroupBox(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_GROUP_BOX, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};

// Message
class wxExtensionMessage: public wxMessage
{
 public:
  wxExtensionMessage(wxPanel *panel, char *label, int x = -1, int y = -1, long style = 0, char *name = "message"):
    wxMessage(panel, label, x, y, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_MESSAGE);
 }
  wxExtensionMessage(wxPanel *panel, wxBitmap *bitmap, int x = -1, int y = -1, long style = 0, char *name = "message"):
    wxMessage(panel, bitmap, x, y, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_MESSAGE);
 }
 ~wxExtensionMessage(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_MESSAGE, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};

// Text
class wxExtensionText: public wxText
{
 public:
  wxExtensionText(wxPanel *panel, wxFunction func, char *label, char *value = "",
         int x = -1, int y = -1, int width = -1, int height = -1, long style = 0, char *name = "text"):
    wxText(panel, func, label, value, x, y, width, height, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_TEXT);
 }

 ~wxExtensionText(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_TEXT, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};

// Multi-line Text
class wxExtensionMultiText: public wxMultiText
{
 public:
  wxExtensionMultiText(wxPanel *panel, wxFunction func, char *label, char *value = "",
         int x = -1, int y = -1, int width = -1, int height = -1, long style = 0, char *name = "multiText"):
    wxMultiText(panel, func, label, value, x, y, width, height, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_MULTI_TEXT);
 }

 ~wxExtensionMultiText(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_MULTI_TEXT, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};

// Check box
class wxExtensionCheckBox: public wxCheckBox
{
 public:
  wxExtensionCheckBox(wxPanel *panel, wxFunction func, char *label,
         int x = -1, int y = -1, int width = -1, int height = -1, long style = 0, char *name = "checkBox"):
    wxCheckBox(panel, func, label, x, y, width, height, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_CHECK_BOX);
 }

 ~wxExtensionCheckBox(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_CHECK_BOX, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};

// Slider
class wxExtensionSlider: public wxSlider
{
 public:
  wxExtensionSlider(wxPanel *panel, wxFunction func, char *label,
         int value, int min_value, int max_value, int width, int x = -1, int y = -1, long style = 0, char *name = "slider"):
    wxSlider(panel, func, label, value, min_value, max_value, width, x, y, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_SLIDER);
 }

 ~wxExtensionSlider(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_SLIDER, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};

// List box
class wxExtensionListBox: public wxListBox
{
 public:
  wxExtensionListBox(wxPanel *panel, wxFunction func, char *label,
                 Bool multiple = FALSE, int x = -1, int y = -1, int width = -1, int height = -1,
                 int noStrings = 0, char **strings = NULL, long style = 0, char *name = "listBox"):
    wxListBox(panel, func, label, multiple, x, y, width, height, noStrings, strings, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_LIST_BOX);
 }

 ~wxExtensionListBox(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_LIST_BOX, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};

// Choice
class wxExtensionChoice: public wxChoice
{
 public:
  wxExtensionChoice(wxPanel *panel, wxFunction func, char *label,
                 int x = -1, int y = -1, int width = -1, int height = -1, int n = 0, char **choices = NULL,
                 long style = 0, char *name = "choice"):
    wxChoice(panel, func, label, x, y, width, height, n, choices, style, name)
 {
   wxStoreThingInTable((long)this, this, wxTYPE_CHOICE);
 }

 ~wxExtensionChoice(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_CHOICE, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};

// RadioBox
class wxExtensionRadioBox: public wxRadioBox
{
public:
    wxExtensionRadioBox(wxPanel *panel, wxFunction func, char *label,
                        int x = -1, int y = -1, int width = -1,
                        int height = -1, int n = 0, char **choices = NULL,
                        int majorDim = 0, long style = 0, char *name = "radioBox"):
    wxRadioBox(panel, func, label, x, y, width, height, n, choices,
               majorDim, style, name)
    {
        wxStoreThingInTable((long)this, this, wxTYPE_RADIO_BOX);
    }

    ~wxExtensionRadioBox(void)
    {
        wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_RADIO_BOX,
                                                           "OnDelete");
        if (func)
            (* (void (*)(long)) func) ((long)this);
        wxDeleteThingInTable((long)this);
    }
};


class wxExtensionMenuBar: public wxMenuBar
{
 public:
  wxExtensionMenuBar(void)
  {
    wxStoreThingInTable((long)this, this, wxTYPE_MENU_BAR);
  }
  ~wxExtensionMenuBar(void)
  {
     wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_MENU_BAR, "OnDelete");
     if (func)
       (* (void (*)(long)) func) ((long)this);
     wxDeleteThingInTable((long)this);
  }
};

class wxExtensionMenu: public wxMenu
{
 public:
  wxExtensionMenu(char *title, wxFunction func):wxMenu(title, func)
  {
    wxStoreThingInTable((long)this, this, wxTYPE_MENU);
  }
  ~wxExtensionMenu(void)
  {
     wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_MENU, "OnDelete");
     if (func)
       (* (void (*)(long)) func) ((long)this);
     wxDeleteThingInTable((long)this);
  }
};

#if USE_GAUGE
// Gauge
class wxExtensionGauge: public wxGauge
{
 public:
  wxExtensionGauge(wxPanel *panel, char *label, int range, int x = -1, int y = -1,
           int width = -1, int height = -1, long style = 0, char *name = "gauge"):
    wxGauge(panel, label, range, x, y, width, height, style, name)
 {
    wxStoreThingInTable((long)this, this, wxTYPE_GAUGE);
 }
 ~wxExtensionGauge(void)
 {
   wxExtensionFunction func = wxFindExtensionCallback(wxTYPE_GAUGE, "OnDelete");
   if (func)
     (* (void (*)(long)) func) ((long)this);
   wxDeleteThingInTable((long)this);
 }
};
#endif

#endif
  // exitemsp.h
  